/* ***********************************************************************
/* File is distrubuted under GPL license, see LICENSE file for datails. /*
/* Author: Lukasz Wojtow <lw@ftw.zamosc.pl>				/*
/* Interesting ideas & some code: Rafal Wijata <www.wijata.com>		/*
/* Dedicate to: Creed							/*
/************************************************************************/

#define CORE_PRIVATE
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
 
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
 
#include "config.h"
#include "version.h"

#define DYNAMIC -1	// dynamic config, allowed only for ~user requests
#define PGCLEAR	1	//PrivsGroups Leave|Clear|Init
#define PGINIT	2
#define OK 0
#define ERR 1

//all per child, used to detect JumpVisits
static char *last_servname = NULL;
static char *last_username = NULL;
static char *curr_username = NULL;
static char dummy_username[] = "-";

#if ISDEBUG==1
//warning at format string, use this with care
#define DEBUG(MSG...) ap_log_rerror(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,r,MSG)
#endif
 
#if ISDEBUG==0
#define DEBUG(MSG...)
#endif

#define LOG_ERROR(MSG...) ap_log_rerror(APLOG_MARK,APLOG_ALERT,r,MSG)
#define LOG_NOERROR(MSG...) ap_log_rerror(APLOG_MARK,APLOG_NOERRNO,r,MSG)
#define ESS 255 // error string size

module MODULE_VAR_EXPORT diffprivs_module;

typedef struct {
   uid_t	uid;		// uid for setuid() // ~user 
   gid_t	gid;		// gid for setgid() // ~user
   char		*username;	// cache for initgroups
   char		*chrootdir;	// dir for chroot   // ~user
   char		*relative;	// dir for relative chroot // ~user
} t_user;

typedef struct {		// see t_user
   uid_t	uid;
   gid_t	gid;
   char		*username;
   char		*chrootdir;
} t_host;

// configuration record for module
typedef struct {
   t_host	*host;		// struct with host requests config
   t_user	*user;		// struct with user requests config
   pool		*mypool;	// if we need more alloc
   unsigned int didall;		// did all or not?
   unsigned int douser;		// do ~user privs?
   unsigned int setgroups;	// call setgroup(0,NULL)?
   unsigned int anypres;	// is any directive present or not
} t_cfg;

// add info to server version
static void mod_diffprivs_init(server_rec *s,pool *p) {
   ap_add_version_component("diffprivs/"VERSION);
   return;
};

// create place for (this virtual) server config
static void *mod_diffprivs_server_config(pool *p, server_rec *s) {
t_cfg *cfg;
   cfg = (t_cfg *) ap_palloc(p,sizeof(t_cfg));
   cfg->host= (t_host *) ap_palloc(p,sizeof(t_host));
   cfg->user= (t_user *) ap_palloc(p,sizeof(t_user));

   cfg->mypool=p;
   cfg->anypres=0;
   cfg->douser=0;
   cfg->setgroups=0;		//default becouse of back.comp.
   cfg->didall=0;
   cfg->user->uid=(uid_t)0;
   cfg->user->gid=(gid_t)0;
   cfg->user->username=dummy_username;
   cfg->user->chrootdir=NULL;
   cfg->user->relative=NULL;
   cfg->host->uid=(uid_t)0;
   cfg->host->gid=(gid_t)0;
   cfg->host->username=dummy_username;
   cfg->host->chrootdir=NULL;
   return (void *) cfg;
}

static int try_cleargroups(request_rec *r) {
	if (setgroups(0,NULL) < 0) {
		LOG_ERROR("euid %u cannot call setgroups",geteuid());
		return ERR;
	};
	return OK;
}
static int try_initgroups(request_rec *r, const char *who, gid_t group) {
	if (initgroups(who,group) < 0) {
		LOG_ERROR("euid %u cannot call initgroups(%s)",geteuid(), who);
		return ERR;
	};
	return OK;
}

static int try_setuid(request_rec *r, uid_t uid) {
	if(uid<MIN_UID) {
		LOG_NOERROR("uid %u is less than minimal allowed uid %d",
				uid,MIN_UID);
		return ERR;
	}
	if(setuid(uid)!=0) {
		LOG_ERROR("euid %u cannot call setuid(%d)",geteuid(),uid);
		return ERR;
	}
	return OK;
}

static int try_setgid(request_rec *r, gid_t gid) {
	if(gid<MIN_GID) {
		LOG_NOERROR("gid %u is less than minimal allowed gid %d",
				gid,MIN_GID);
		return ERR;
	}
	if(setgid(gid)!=0) {
		LOG_ERROR("euid %u cannot call setgid(%d)",geteuid(),gid);
		return ERR;
	}
	return OK;
}

static int try_chroot(request_rec *r, char *dir) {
	DEBUG("try_chroot() to dir %s",dir);
	// checks was already done while reading config.
	// if dynamic dir is used, it will match criteria
	if(chroot(dir)!=0) {
		LOG_ERROR("euid %u cannot call chroot(%s)",geteuid(),dir);
		return ERR;
	}
	if(chdir("/")!=0) {
		LOG_ERROR("euid %d cannot call chdir(/) after chroot(%s)",
			     geteuid(),dir);
		return ERR;
	}
	return OK;
};

static int do_user(request_rec *r, t_cfg *cfg) {
	uid_t uid;
	gid_t gid;
	struct passwd *pwd=NULL;
	char *chrootdir=NULL;
	char path[PATH_MAX]={'\0'};
	char real_path[PATH_MAX]={'\0'};

	DEBUG("do_user() for uri: %s",r->uri);
	// if dynamic chroot, then chrootdir to user's home, 
	// else cfg->user->chrootdir, dont do chroot yet
	chrootdir=cfg->user->chrootdir;
	if (chrootdir == (char*)DYNAMIC) {
	    if (!(pwd = getpwnam(curr_username))) {
		LOG_ERROR("getpwnam(%s) failed", curr_username);
		return HTTP_SERVICE_UNAVAILABLE;
	    };
	    chrootdir = pwd->pw_dir;
	    if(cfg->user->relative) {
	    // if relative chroot then, we have to get full path for this relative
	    // dir by calling realpath()
		ap_snprintf(path,sizeof(path),"%s/%s",
			chrootdir,cfg->user->relative);
		DEBUG("Doing relative chroot to %s",path);
		if(realpath(path,real_path)==NULL) {
			LOG_ERROR("Cannot get realpath for %s",path);
			return HTTP_SERVICE_UNAVAILABLE;
		}
		chrootdir=real_path;
	    }
	};//if DYNAMIC
	// first we should try cleargroups
	if(cfg->setgroups==PGCLEAR && try_cleargroups(r))
		return HTTP_SERVICE_UNAVAILABLE;
	// if ?ID dynamic, grab pwdata before we chroot
	if (cfg->user->uid==DYNAMIC || cfg->user->gid==DYNAMIC) {
		if (!pwd) pwd = getpwnam(curr_username);//no error loging yet
		// initgroups should be called where user was found
		if (pwd && cfg->setgroups==PGINIT &&
		    try_initgroups(r, pwd->pw_name, pwd->pw_gid))
			return HTTP_SERVICE_UNAVAILABLE;
	} else {
		if (cfg->setgroups==PGINIT &&
		    try_initgroups(r, cfg->user->username, cfg->user->gid))
			return HTTP_SERVICE_UNAVAILABLE;
	};
	//chroot
	if(chrootdir && try_chroot(r, chrootdir))
		return HTTP_SERVICE_UNAVAILABLE;
	//if ?ID dynamic, and stiil no pwd, try grab it again
	if (cfg->user->uid==DYNAMIC || cfg->user->gid==DYNAMIC) {
	    if (!pwd && !(pwd = getpwnam(curr_username))) {
		    LOG_ERROR("getpwnam(%s) failed", curr_username);
		    return HTTP_SERVICE_UNAVAILABLE;
	    };
	    //we have pwd now
	    if (cfg->setgroups==PGINIT &&
		try_initgroups(r, pwd->pw_name, pwd->pw_gid))
		    return HTTP_SERVICE_UNAVAILABLE;
	};
	// if dynamic config, then *id as user *id, else as cfg->user_*id
	uid=cfg->user->uid==DYNAMIC?pwd->pw_uid:cfg->user->uid;
	gid=cfg->user->gid==DYNAMIC?pwd->pw_gid:cfg->user->gid;
	//setgid
	if(gid && try_setgid(r, gid))
		return HTTP_SERVICE_UNAVAILABLE;
	//setuid
	if(uid && try_setuid(r, uid)) 
		return HTTP_SERVICE_UNAVAILABLE;
	cfg->didall=1;
	return DECLINED;
}

static int do_non_user(request_rec *r, t_cfg *cfg)
{
	DEBUG("do_non_user: chrootdir: %s, gid: %d, uid %d",
		cfg->host->chrootdir,cfg->host->gid,cfg->host->uid);
	//no dynamic configs, rather simple
	if(cfg->setgroups==PGCLEAR && try_cleargroups(r))
		return HTTP_SERVICE_UNAVAILABLE;
	if(cfg->setgroups==PGINIT &&
	    try_initgroups(r, cfg->host->username,
			      cfg->host->gid ? cfg->host->gid : getegid()))
		return HTTP_SERVICE_UNAVAILABLE;
	if(cfg->host->chrootdir
		&& try_chroot(r, cfg->host->chrootdir))
		return HTTP_SERVICE_UNAVAILABLE;
	if(cfg->host->gid && try_setgid(r, cfg->host->gid))
		return HTTP_SERVICE_UNAVAILABLE;
	if(cfg->host->uid && try_setuid(r, cfg->host->uid)) 
		return HTTP_SERVICE_UNAVAILABLE;
	cfg->didall=1;
	return DECLINED;
}

static int did_jump(request_rec *r) {
    LOG_NOERROR("JumpVisit detected %s > %s : %s > %s",
		last_servname, r->hostname, last_username, curr_username);
    //let client know, that this error is really temporary
    if (r->status == HTTP_OK) {	//status not really set yet
	//from RFC, but not working
	ap_table_setn(r->err_headers_out, "Retry-After", "0");
	//this will actually work in IE/Netscape/Mozilla
	ap_table_setn(r->err_headers_out, "Refresh", "0");
    };
    //BTW, following error closes connection, but child
    //is not required to exit. However if MaxKeepAliveRequests is set to 1
    //child will exit
    return HTTP_SERVICE_UNAVAILABLE;
};

static int mod_diffprivs_first(request_rec *r) {
t_cfg	*cfg;
char	*ptr;

	DEBUG("Called mod_diffprivs_first");
	cfg=ap_get_module_config(r->server->module_config,&diffprivs_module);
	// if nothing to work, then back
	if(! cfg->anypres) return DECLINED;
	// if ~user request
	if(r->uri[0]=='/' && r->uri[1]=='~') {
		if (! cfg->douser) {
			LOG_NOERROR("%s%s request is not allowed by diffprivs",
				    r->hostname, r->uri);
			return HTTP_FORBIDDEN;
		};
		//jump server->user user1->user2
		if (last_servname)
			return did_jump(r);
		//get username, TODO: free after ?
		if (curr_username) free(curr_username);
		curr_username = strdup(r->uri+2);
		if (ptr = strchr(curr_username, '/')) *ptr = 0;
		if (last_username==NULL) last_username = strdup(curr_username);
		if (strcmp(last_username, curr_username))
			return did_jump(r);
		if (cfg->didall) return DECLINED;
		return do_user(r,cfg);
	} else {// if not ~user request then do normal work
		//jump user->server server1->server2
		if (last_username)
			return did_jump(r);
		if (last_servname==NULL && r->hostname)
			last_servname=strdup(r->hostname);
		if (r->hostname && strcasecmp(last_servname,r->hostname))
			return did_jump(r);
		if (cfg->didall) return DECLINED;
		return do_non_user(r,cfg);
	};
}

static int parse_id(int isuid, uid_t *id, const char *idname, char *error) {
char *endptr=NULL;
struct passwd *pwd=NULL;
struct group *grp=NULL;

	//do nothing
	if(idname==NULL || !strcasecmp("DoNothing",idname)) {
		*id=(uid_t)0;
		return OK;
	};
	// dynamic config
	if(isuid ? !strcasecmp("AsUserUid",idname)
	         : !strcasecmp("AsUserGid",idname)) {
		*id=DYNAMIC;
		return OK;
	};
	// idname given as #id
	if(*idname=='#') {
		*id=strtol(idname+1,&endptr,10);
		// if strtol return error
		if(endptr && *endptr) {
			ap_snprintf(error,ESS,"Invalid ID: %s",idname);
			return ERR;
		}
		return OK;
	};
	// user/group given as name
	if (isuid) {
		pwd=getpwnam(idname);
		if(pwd==NULL) {
			ap_snprintf(error,ESS, "Invalid user: %s",idname);
			return ERR;
		}
		*id=pwd->pw_uid;
		if(*id<MIN_UID) {
		    ap_snprintf(error,ESS,"Miminal allowed uid is %d, but used %d "
				"for %s",MIN_UID,*id,idname);
		    return ERR;
		}
	} else {
		grp=getgrnam(idname);
		if(grp==NULL) {
			ap_snprintf(error,ESS, "Invalid group: %s",idname);
			return ERR;
		}
		*id=grp->gr_gid;
		if(*id<MIN_GID) {
		    ap_snprintf(error,ESS,"Miminal allowed gid is %d, but used %d "
				"for %s",MIN_GID,*id,idname);
		    return ERR;
		}
	};
	return OK;
}

static int parse_dir(char **chrootdir,const char *dir,char *error) {
	// if no chroot
	if(dir==NULL) {
		*chrootdir=NULL;
		return OK;
	}
	// if dynamic config
	if(!strncasecmp("AsUserHome",dir,10)) {
		*chrootdir=(char *)DYNAMIC;
		return OK;
	}
	// if dir not exist
	if(access(dir,F_OK)!=0) {
		ap_snprintf(error,ESS,"File not found: %s",dir);
		return ERR;
	}
	// if it's under UNDER_HERE value from config.h
	if(strncmp(UNDER_HERE,dir,strlen(UNDER_HERE))) {
		ap_snprintf(error,ESS,
			"Cannot do chroot for dir %s (it is not under %s)",
			dir,UNDER_HERE);
		return ERR;
	}
	// don't want relative paths as directory
	if(strstr(dir,"..")) {
		ap_snprintf(error,ESS,
			"Not allowed characters in path: %s",dir);
		return ERR;
	}
	// as above
	if(dir[0]!='/') {
		ap_snprintf(error,ESS,"Path must be non-relative: %s",dir);
		return ERR;
	}
	*chrootdir=(char *)dir;
	return OK;
}

// get Privs argument from config
static const char *privs(cmd_parms *cmd, void *mconfig,
                             const char *user,
                             const char *group,
                             const char *dir) {
static char error[ESS];
struct passwd *pwd;
// get config
t_cfg *cfg = ap_get_module_config(cmd->server->module_config,
						&diffprivs_module);
	memset(error,0,ESS);
	if(parse_id(1, &cfg->host->uid,user,error))
		return error;
	if(cfg->host->uid > 0 && (pwd = getpwuid(cfg->host->uid)))
		cfg->host->username = ap_pstrdup(cfg->mypool, pwd->pw_name);
	if(parse_id(0, &cfg->host->gid,group,error))
		return error;
	if(parse_dir(&cfg->host->chrootdir,dir,error))
		return error;
	// dynamic config can be only for PrivsUser directive, not here
	if(	cfg->host->uid==DYNAMIC || 
		cfg->host->gid==DYNAMIC || 
		cfg->host->chrootdir==(char *)DYNAMIC) {
		ap_snprintf(error,ESS, "Cannot do dynamic config in Privs");
		return error;
	}
	cfg->anypres=1;
	return NULL;
};

static const char *privsuser(cmd_parms *cmd,void *mconfig,
				const char *user,
				const char *group,
				const char *dir)
{
static char error[ESS];
struct passwd *pwd;
// get config
t_cfg *cfg = ap_get_module_config(cmd->server->module_config,
						&diffprivs_module);
	memset(error,0,ESS);
	if(parse_id(1, &cfg->user->uid,user,error))
		return error;
	if(cfg->user->uid > 0 && (pwd = getpwuid(cfg->user->uid)))
		cfg->user->username = ap_pstrdup(cfg->mypool, pwd->pw_name);
	if(parse_id(0, &cfg->user->gid,group,error))
		return error;
	if(parse_dir(&cfg->user->chrootdir,dir,error))
		return error;
	// if there is something after AsUserHome then set it as relative
	if(cfg->user->chrootdir==(char *)DYNAMIC && strlen(dir)>strlen("AsUserHome"))
		cfg->user->relative=(char *)dir+strlen("AsUserHome");
	// both uid/gid must be dynamic or not.
	if (  (cfg->user->uid==DYNAMIC || cfg->user->gid==DYNAMIC)
	    &&(cfg->user->uid!=cfg->user->gid) ) {
		ap_snprintf(error,ESS, "Both uid & gid must be DYNAMIC or NOT");
		return error;
	};
	cfg->douser=1;
	cfg->anypres=1;
	return NULL;
}

static const char *privsgroups(cmd_parms *cmd, void *mconfig, const char *how)
{
static char error[ESS];
// get config
t_cfg *cfg = ap_get_module_config(cmd->server->module_config,
						&diffprivs_module);
	memset(error,0,ESS);
	if (!strcasecmp("Leave",how)) return NULL; // Leave has no efect
	else if (!strcasecmp("Off",how)) return NULL; //backward compatibility
	else if (!strcasecmp("Clear",how)) cfg->setgroups=PGCLEAR;
	else if (!strcasecmp("On",how)) cfg->setgroups=PGCLEAR; //backcomp
	else if (!strcasecmp("Init",how)) cfg->setgroups=PGINIT;
	else {
		ap_snprintf(error,ESS,"Unknown: PrivsGroups %s", how);
		return error;
	};
	cfg->anypres=1;
	return NULL;
}

// table of commands
static const command_rec mod_diffprivs_commands[] =
{
	{
	"Privs", privs, NULL, RSRC_CONF, TAKE123,
		"Diffrent process privileges: \"Privs uid [gid [chroot_dir]]\""
	},

	{
	"PrivsUser",privsuser,NULL,RSRC_CONF,TAKE123,
		"Diffrent privileges for user's "
			"(www.server.com/~user) requests: \""
			"PrivsUser uid [gid [chroot_dir]]\""
	},
	{
	"PrivsGroups",privsgroups,NULL,RSRC_CONF,TAKE1,
		"Call or not setgroups(): \"PrivsGroups Leave|Clear|Init\""
	},
     {NULL}
};

module MODULE_VAR_EXPORT diffprivs_module =
{
   STANDARD_MODULE_STUFF,
     mod_diffprivs_init,                // add info about module
     NULL, NULL,
     mod_diffprivs_server_config,       // create server config
     NULL,				// server merger
     mod_diffprivs_commands,            // module commands in httpd.conf
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     mod_diffprivs_first                // when server read request.
};
